*8290.6 GST 229 Sort Module
*
*  This is a reasonably general SORT subroutine.
*
*  The input data structure consists of a set of contiguous variable-
*length records with single-byte delimiters in a RAM buffer.
*
*  The algorithm is a Shell sort (after Donald W. Shell) via detached
*keys.  A version of Shell sort and an explanation thereof is to be
*found in many text books of computer science.
*
*  The comparison routine (to compare whole records) is passed as a
*parameter to this routine so that the algorithm is independent of the
*collating sequence.
*
*  This version calls two externally defined routines "MEMGET" and
*"MEMDROP" to grab and release memory (other than stack space) required
*for workspace.
*
*  This module is provided as part of the example program.
*
         INCLUDE  MDV2_MACRO_LIB
         PAGE
         XDEF     SORT              Declare entry point for linker.

*        Registers   On entry                On exit
*
*        D0.L        No. of records          preserved
*        D1.B        delimiter code          preserved
*        A0.L        pointer to buffer       preserved (buffer updated)
*        A1.L        comparison routine      preserved
*        others      don't care              preserved

*This routine also uses ...

         XREF     MEMGET            ... whose interface is as follows

*        Registers   On entry                On exit
*
*        D0.L        no. of bytes required   control data for MEMDROP
*        A0.L        undefined               address of obtained RAM
*        others      undefined               must be preserved

* ... and ...

         XREF     MEMDROP           ... whose interface is ...

*        Registers  On entry                  On exit
*
*        D0         control data from MEMGET  don't care
*        A0         address of memory         don't care
*        others     undefined                 must be preserved

*The comparison routine whose address is given in A1 on entry must have
*the following interface ...

*        Registers   On entry                On exit
*
*        D0.L        undefined               result of comarison
*        A0.L        address of 1st record   must be preserved
*        A1.L        address of 2nd record   must be preserved
*        others      undefined               must be preserved

* ... where "result of comparison" is defined as follows ...
*
*        D0 < 0  means  A0 points to record EARLIER in collating sequence
*        D0 = 0  means  A0 and A1 point to identical records
*        D0 > 0  means  A0 points to record  LATER  in collating sequence
         PAGE
*Stack record layout definition:
*
*This set of offsets is used to define the contents of a small area of
*RAM grabbed from the top of the stack at the start of this routine and
*used for temporary storage.

         OFFSET   0                 Stack record layout definition

NRECS    DS.L     1                 No. of records to sort
GIVENBUF DS.L     1                 Address of buffer passed by caller
GOTBUF   DS.L     1                 Address of similar (grabbed) buffer
PTRBUF   DS.L     1                 Address of buffer for detached keys
COMPARE  DS.L     1                 Comparison routine entry point
DELIM    DS.B     1                 Record delimiter byte

STRECLEN DS.W     0                 STRECLEN = length of stack record

*The last equivalence to declare is the "step factor".  This determines
*how quickly the interval between comparison points approaches one.
*The smallest value for this constant which will work is two, the value
*three here is based on a recommendation by Maurice V. Wilkes.

STEPFAC  EQU      3                 See above.

         SECTION  CODE              Every module needs something like this

*The logic starts here.  First set up stack record.

         SUBROUTINE$ SORT,D0-D7/A0-A4/A6

         SUB.L    #STRECLEN,A7      Set upstack record
         MOVE.L   A7,A6             Keep pointer to it

         MOVE.L   D0,NRECS(A6)      No. of records to sort
         MOVE.B   D1,DELIM(A6)      Delimiter byte
         MOVE.L   A0,GIVENBUF(A6)   where the data to be sorted is
         MOVE.L   A1,COMPARE(A6)    Record comparison routine address
         PAGE
*Next step is to grab the necessary workspace.

*The size of the pointer buffer is easily calculated from the number
*of records to be sorted - one longword for each record.

         ASL.L    #2,D0             Multiply by four
         CALL$    MEMGET            Grab that much memory
         MOVE.L   A0,PTRBUF(A6)     Remember where it is

*To determine ow much memory is needed for the other workspace buffer
*must look at all the records.  At the same time, can extract the
*pointers to their starts and put them in the pointer buffer for use
*as detached keys.

         MOVE.L   GIVENBUF(A6),A1   where the records are
         MOVE.L   NRECS(A6),D0      how many of them there are
         MOVE.B   DELIM(A6),D1      get delimiter byte

         WHILE    D0 GT.L #0 DO.S   for each record ...
            MOVE.L   A1,(A0)+          put its address into the ptr. buf.
            WHILE    (A1)+ NE.B D1 DO.S
            ENDWHILE                   find start of next record
            SUBQ.L   #1,D0          keep count
         ENDWHILE

         SUB.L    GIVENBUF(A6),A1   Calculate relevant length of buffer
         MOVE.L   A1,D0             That's how many bytes to grab
         CALL$    MEMGET            Grab second working buffer
         MOVE.L   A0,GOTBUF(A6)     and buffer address

         PAGE
*Now all the data has been set up start the sort.  This section is less
*well commented than the rest of the module since it is just an
*implementation of a standard algorithm.

         MOVE.L   NRECS(A6),D1      No. of records to sort
         MOVE.L   D1,D2             Initialise increment

         REPEAT
            ADDQ.L   #(STEPFAC-1),D2
            DIVS     #STEPFAC,D2
            EXT.L    D2                Increment for this pass

            MOVE.L   D2,D3
            SUBQ.L   #1,D3             telegraph pole
            ASL.L    #2,D3             times 4 for longwords
            ADD.L    PTRBUF(A6),D3     point to last entry in pointers

            FOR.L    D4 = PTRBUF(A6) TO D3 BY #4 DO

               MOVE.L   D4,D5          first record
               MOVE.L   D1,D6
               SUB.L    D2,D6
               SUBQ.L   #1,D6
               ASL.L    #2,D6
               ADD.L    PTRBUF(A6),D6   last record to compare

               WHILE    D5 LE.L D6 DO

                  MOVE.L   D5,A2             The "bubble sort" part
                  REPEAT

                     MOVE.L   A2,A3
                     ADD.L    D2,A3
                     ADD.L    D2,A3
                     ADD.L    D2,A3
                     ADD.L    D2,A3
                     MOVE.L   (A3),A1           Get a pointer to a record
                     MOVE.L   (A2),A0           and another
                     MOVE.L   COMPARE(A6),A4    Get address of comparison routine
                     JSR      (A4)              Compare them

                     TST.L    D0                Look at result
                     IF       GT THEN.S         If they must be swapped

                        MOVE.L   A0,(A3)           Put pointers back into buffer ...
                        MOVE.L   A1,(A2)           ... the other way round!
                        MOVE.L   D2,D7
                        ASL.L    #2,D7             times 4 for longwords
                        SUB.L    D7,A2             step pointer back

                     ENDIF

                  UNTIL    D0 LE.L #0 OR A2 LT.L D4

                  MOVE.L   D2,D7
                  ASL.L    #2,D7
                  ADD.L    D7,D5

               ENDWHILE

            ENDFOR

         UNTIL    D2 EQ.L #1

         PAGE
*Now the detached keys in the pointer buffer have been sorted, the
*records they point at are copied in the new order into the second
*buffer and then copied back into the original given buffer.

         MOVE.L   NRECS(A6),D0         Number of records to copy
         MOVE.L   GOTBUF(A6),A0        Where to put them
         MOVE.L   PTRBUF(A6),A1        Where the pointers to them are
         MOVE.B   DELIM(A6),D1         The delimiter byte

         WHILE    D0 GT.L #0 DO.S      For each record ...
            MOVE.L   (A1)+,A2             Get base address of that record
            REPEAT                        For each byte in it ...
               MOVE.B   (A2)+,(A0)+          Copy a byte
            UNTIL    -1(A2) EQ.B D1      ;Until delimiter has been copied

            SUBQ.L   #1,D0             Number of bytes left to copy
         ENDWHILE

         MOVE.L   A0,D0                Pointer to logical end of 2nd buffer
         MOVE.L   GOTBUF(A6),A0        Pointer to start of same
         MOVE.L   GIVENBUF(A6),A1      Pointer to start of given buffer

         REPEAT
            MOVE.B   (A0)+,(A1)+          Copy ...
         UNTIL    A0 EQ.L D0           ;        ... the whole buffer

*Finally release the temporary storage and return to caller

         MOVE.L   GOTBUF(A6),A0
         CALL$    MEMDROP                 Release second buffer

         MOVE.L   PTRBUF(A6),A0
         CALL$    MEMDROP                 Release pointer buffer

         ADD.L    #STRECLEN,A7            Release stack record

         END$     SORT                    Restore registers and return
         END

